/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <ncurses.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "main.h"

#include "curs_color.h"
#include "lib.h"
#include "misc.h"
#include "options.h"

#define USE_ERRNO 1
#define NO_ERRNO  0

static void init_homedir(void);
static int source_rc(const char *rcfile);
static int parse_rc_line(char *s, char *errmsg, size_t errlen,
			 const char *rcfile);
static void hit_enter_to_continue(bool use_errno, const char *fmt,...);

void init_curses(void)
{
    initscr(); /* Initialize curses, exit on error */
    init_colors(); /* Initialize colors, exit on errror */
    /* Do not change these! */
    keypad(stdscr, TRUE); /* Enable terminal's keypad */
    cbreak();
    noecho();
}

void init_vars(void)
{
    /* initalize HomeDir */
    init_homedir();

    if (access(GLOBAL_RC_FILE, F_OK) == 0)
	(void) source_rc(GLOBAL_RC_FILE);	/* do nothing if the file doesn't exist */

    PersonalRc = safe_malloc(strlen(HomeDir) + strlen(RC_FILENAME) + 2);
    sprintf(PersonalRc, "%s/%s", HomeDir, RC_FILENAME);
    if (access(PersonalRc, F_OK) == 0) {	/* if it exists */
	if (source_rc(PersonalRc) != 0)		/* then source it */
	    exit(EXIT_FAILURE);
    }

    initialize_unset_options();

    if (is_string_option_set(OPT_FILENAME_EXTENSIONS)) {
	if (has_the_correct_syntax_of_the_filename_extensions_variable(string_option_value(OPT_FILENAME_EXTENSIONS)) == FALSE) {
	    printf("Error: \"%s\" does not match the syntax of the %s variable.\n"
		   "Edit it through %s or %s (the syntax is explained there).\n",
		   string_option_value(OPT_FILENAME_EXTENSIONS),
		   option_name(OPT_FILENAME_EXTENSIONS), GLOBAL_RC_FILE, PersonalRc);
	    exit(1);
	}
    }

    if (is_path_option_set(OPT_STARTUP_WORKING_DIR)) {
	if (xchdir(path_option_value(OPT_STARTUP_WORKING_DIR)) != 0) {
	    printf("Can't change directory to %s (specified in the %s variable "
		   "in %s or %s): %s\n.", path_option_value(OPT_STARTUP_WORKING_DIR),
		    option_name(OPT_STARTUP_WORKING_DIR), PersonalRc, GLOBAL_RC_FILE,
		    strerror(errno));
	    exit(1);
	}
    } else {
        char path[PATH_MAX + 1];
	if (getcwd(path, sizeof(path)) == NULL) {
	    perror("Can't get current working directory");
	    exit(1);
	}
	if (xchdir(path) != 0) {
	    printf("Can't change directory to %s: %s (errno = %d)\n",
		   path, strerror(errno), errno);
	    exit(1);
	}
    }
}

/*
 * Initalize HomeDir using the environment variable HOME, if not found, use
 * pw_dir from the passwd entry. Exit program on error.
 */
static void init_homedir(void)
{
    struct passwd *pwd;
    char *p, *q;

    p = getenv("HOME");
    if (p && *p) {
	if (*p != '/') {
	    printf("Environment variable HOME does not start with slash: %s.\n", p);
	    exit(1);
	}
	if (strcmp(p, "/") != 0) {	/* if p is not the root directory */
	    q = p + strlen(p) - 1;
	    /* remove any trailing slashes */
	    while (*q == '/' && q > p)
		*q-- = '\0';	/* we must not use HOME again since this will change its value */
	    /* p does not end with slash now */
	}
	HomeDir = safe_strdup(p);

    } else {
	pwd = getpwuid(getuid());
	if (!pwd) {
	    perror("Can't get your password file entry (which is needed to get your "
		   " home directory when the environment variable HOME is not set)");
	    exit(EXIT_FAILURE);
	}
	if (!pwd->pw_dir) {
	    printf("Can't find your home directory: environment variable HOME "
		   " not set and no home directory entry found in the passwd file.");
	    exit(EXIT_FAILURE);
	}
	if (strcmp(pwd->pw_dir, "/") != 0) {	/* if pwd->pw_dir is not the root directory */
	    q = pwd->pw_dir + strlen(pwd->pw_dir) - 1;
	    /* remove any trailing slashes */
	    while (*q == '/')
		*q-- = '\0';
	    /* pwd->pw_dir does not end with slash now */
	}
	HomeDir = safe_strdup(pwd->pw_dir);
    }

    /* W_OK is really only needed if you need to save the options. */
    if (access(HomeDir, R_OK | W_OK | X_OK) != 0) {
	printf("Can't access your home directory %s: %s.\n", HomeDir, strerror(errno));
	exit(1);
    }
}

static int source_rc(const char *rcfile)
{
    FILE *fp;
    int line = 0;
    int isContinuation = 0;
    char buf[LONG_STRING];
    char errbuf[SHORT_STRING];
    char *linebuf = NULL;
    size_t linelen = 0;
    size_t linemax = 0;
    size_t buflen;

    if ((fp = fopen(rcfile, "r")) == NULL) {
	char *buf = safe_malloc(strlen(rcfile) + strlen(": unable to open file") + 1);	/* i'm too lazy to count */
	strcpy(buf, rcfile);
	strcat(buf, ": unable to open file");
	perror(buf);
	return -1;
    }
    buf[sizeof(buf) - 1] = '\0';
    while (fgets(buf, sizeof(buf) - 1, fp)) {
	line++;

	buflen = strlen(buf);

	/* check to see if the whole line was read */

	isContinuation = 0;
	if (buf[buflen - 1] == '\n') {
	    buflen--;
	    buf[buflen] = '\0';	/* kill the newline */

	    /* look for a continuation line */
	    if (buflen && buf[buflen - 1] == '\\') {
		buflen--;
		isContinuation = 1;
	    }
	} else
	    isContinuation = 1;

	if (buflen) {
	    if (buflen + linelen > linemax) {
		/* not enough room, so allocate more */
		safe_realloc((void **) &linebuf, (linemax = buflen + linelen + 80));
	    }
	    memcpy(linebuf + linelen, buf, buflen + 1);
	    linelen += buflen;

	    if (isContinuation)
		continue;
	}
	if (linelen) {
	    if (parse_rc_line(linebuf, errbuf, sizeof(errbuf), rcfile) == -1)
		hit_enter_to_continue(NO_ERRNO, "Error in %s, line %d: %s", rcfile, line, errbuf);
	    linelen = 0;
	}
    }

    safe_free(linebuf);
    (void) fclose(fp);

    return 0;
}

static int parse_rc_line(char *s, char *errmsg, size_t errlen, const char *rcfile)
{
    char *cp, *cp2, *left, *right;
    int option_index;

    if ((cp = strchr(s, '#')))	/* is there a comment on this line */
	*cp = '\0';
    if ((right = strchr(s, '=')) == NULL) {
	cp2 = s;
	while (isspace((unsigned char) *cp2))
	    cp2++;
	if (*cp2 && cp == NULL) {	/* if there were stuff on this line, but no comment or = */
	    snprintf(errmsg, errlen, "syntax error, no assignment found: \"%s\".", s);
	    return -1;
	}
	return 0;
    }
    *right++ = '\0';
    left = s;

    if (!(left = rm_leading_and_trailing_ws(left))) {
	snprintf(errmsg, errlen, "syntax error, no assignment found.");
	return -1;
    }
    if ((option_index = get_option_index_by_name(left)) == -1) {
	snprintf(errmsg, errlen, "unknown variable: \"%s\"", left);
	return -1;
    }
    if (!(right = rm_leading_and_trailing_ws(right)))
	return 0;		/* if right is "\0" or only contains whitespace */

    /* If the option is already set, we overwrite the old value. */
    if (is_a_boolean_option(option_index)) {
	if (strcasecmp(right, "yes") == 0 || strcasecmp(right, "true") == 0 ||
	    strcasecmp(right, "on") == 0 || strcasecmp(right, "1") == 0)
	    set_boolean_option(option_index, TRUE);
	else if (strcasecmp(right, "no") == 0 || strcasecmp(right, "false") == 0 ||
	      strcasecmp(right, "off") == 0 || strcasecmp(right, "0") == 0)
	    set_boolean_option(option_index, FALSE);
	else {
	    snprintf(errmsg, errlen, "\"%s\" must be one of: yes/no/true/false/on/off.", option_name(option_index));
	    return -1;
	}
	mark_option_as_initialized(option_index);
	return 0;
    } else if (is_a_path_option(option_index)) {
	char buf[PATH_MAX + 1];
	char *p = NULL;
	int saved_errno = 0;
	struct stat statbuf;

	if (right[0] != '/' && right[0] != '~') {
	    fprintf(stderr, "Path variable %s in %s does not start with / or ~: \"%s\".\n",
		    option_name(option_index), rcfile, right);
	    exit(EXIT_FAILURE);
	}
	strfcpy(buf, right, sizeof(buf));
	if (expand_path(buf, sizeof(buf)) == NULL) {
	    fprintf(stderr, "Can't expand the path variable %s in %s: \"%s\".\n",
		    option_name(option_index), rcfile, right);
	    exit(EXIT_FAILURE);
	}

	if (stat(buf, &statbuf)) {
	    saved_errno = errno;
	    p = strerror(saved_errno);
	    fprintf(stderr, "Can't stat %s \"%s\" specified in %s: %s (errno = %d).\n",
		    option_name(option_index), right, rcfile, p ? p : "unknown error", saved_errno);
	    exit(EXIT_FAILURE);
	}

		  /* If we don't have read permission */
	if ((!(statbuf.st_mode & S_IRUSR)) ||
	    /* or if it's a directory and we don't have execute permission. */
	    (S_ISDIR(statbuf.st_mode) && !(statbuf.st_mode & S_IXUSR))) {
	    saved_errno = errno;
	    p = strerror(saved_errno);
	    fprintf(stderr, "Can't access %s \"%s\" specified in %s: %s (errno = %d).\n",
		    option_name(option_index), right, rcfile, p ? p : "unknown error", saved_errno);
	    exit(EXIT_FAILURE);
	}
	set_path_option(option_index, buf);
	mark_option_as_initialized(option_index);
    } else if (is_a_string_option(option_index)) {
	set_string_option(option_index, right);
	mark_option_as_initialized(option_index);
    } else if (is_a_number_option(option_index)) {
	int n = atoi(right);
	if (n >= 0)
	    set_number_option(option_index, n);
	else {
	    snprintf(errmsg, errlen, "\"%s\" must be a value >= 0.", left);
	    return -1;
	}
		mark_option_as_initialized(option_index);
	return 0;
    } else
	assert(0);

    return 0;
}

static void hit_enter_to_continue(bool use_errno, const char *fmt,...)
{
    va_list ap;
    int saved_errno = errno;
    char *p = NULL;

    if (fmt) {
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
    }
    if (use_errno) {
	p = strerror(errno);
	printf("%s%s (errno = %d).", fmt ? "" : ": ", p ? p : "unknown error", saved_errno);
    }
    printf("\nHit Enter to continue...");
    fflush(stdout);

    (void) getchar();		/* does not return before you hit enter */
}
